
-- | Normal form of braids, take 1
--
--
-- Based on:
--
-- Joan S. Birman, Tara E. Brendle: BRAIDS - A SURVEY
-- <https://www.math.columbia.edu/~jb/Handbook-21.pdf>
--


{-# LANGUAGE 
      CPP, BangPatterns, 
      ScopedTypeVariables, ExistentialQuantification,
      DataKinds, KindSignatures, Rank2Types #-}

module BraidNF where

--------------------------------------------------------------------------------

import Data.Proxy
import GHC.TypeLits

import Control.Monad

import Data.List ( mapAccumL , foldl' )

import Data.Array.Unboxed
import Data.Array.ST
import Data.Array.IArray
import Data.Array.MArray
import Data.Array.Unsafe
import Data.Array.Base

import Control.Monad.ST

import Math.Combinat.Helper
import Math.Combinat.Sign

import Math.Combinat.Permutations ( Permutation(..) )
import qualified Math.Combinat.Permutations as P

import Groups
import Braid

#ifdef QUICKCHECK
import Test.QuickCheck
#endif

--------------------------------------------------------------------------------
{- * The word problem in the braid group, take 1

We implement the Adyan-Thurston-ElRifai-Morton solution to the word problem,
which should be @O(|w|^2*n*log(n))@ 
(we may be slightly more inefficient because of purity \/ immutable data 
structures, haven\'t really thought about it).

See Chapter 5.1 in:

Joan S. Birman, Tara E. Brendle: BRAIDS - A SURVEY
<https://www.math.columbia.edu/~jb/Handbook-21.pdf>

-}

-- | A unique normal form for braids, called the /left-greedy normal form/,
-- looks like @Delta^i*P@, where @Delta@ is the positive half-twist, @i@ is an integer,
-- and @P@ is a positive word, which can be further decomposed into non-@Delta@ /permutation words/; 
-- these words themselves are not unique, but the permutations they realize /are/ unique.
--
-- This will solve the word problem relatively fast, 
-- though it is not the fastest known algorithm.
--
data BraidNF (n :: Nat) = BraidNF
  { _nfDeltaExp :: !Int              -- ^ the exponent of @Delta@
  , _nfPerms    :: [Permutation]     -- ^ the permutations
  }
  deriving (Eq,Ord,Show)

-- | A braid word representing the given normal form
nfReprWord :: KnownNat n => BraidNF n -> Braid n
nfReprWord (BraidNF k perms) = freeReduceBraidWord $ composeMany (deltas ++ rest) where

  deltas 
    | k > 0     = replicate   k           halfTwist
    | k < 0     = replicate (-k) (inverse halfTwist)
    | otherwise = []

  rest = map permutationBraid perms

-- | Computes the normal form of a braid. We apply free reduction first, it should be faster that way.
braidNormalForm :: KnownNat n => Braid n -> BraidNF n
braidNormalForm = braidNormalForm' . freeReduceBraidWord

-- | This function does not apply free reduction before computing the normal form
braidNormalForm' :: KnownNat n => Braid n -> BraidNF n
braidNormalForm' braid@(Braid gens) = BraidNF (dexp+pexp) perms where
  n = numberOfStrands braid
  invless = replaceInversesNaive gens
  (dexp,posxword) = moveDeltasLeft n invless
  factors = leftGreedyFactors n $ expandPosXWord n posxword
  -- we should recognize positive deltas here and move them to the left!
  (pexp,factors') = filterDeltaFactors n $ normalizeFactors factors 
  perms   = map (artinWordPermutation n) factors'

--------------------------------------------------------------------------------

-- | Replaces groups of @sigma_i^-1@ generators by @(Delta^-1 * P)@, 
-- where @P@ is a positive word.
--
-- This should be more clever (resulting in shorter words) than the naive version below
--
replaceInverses :: Int -> [BrGen] -> [XGen]
replaceInverses n gens = worker gens where

  worker [] = []
  worker xs = replaceNegs neg ++ map (XSigma . brGenIdx) pos ++ worker rest where 
    (neg,tmp ) = span (isMinus . brGenSign) xs
    (pos,rest) = span (isPlus  . brGenSign) tmp
  
  replaceNegs gs = concatMap replaceFac facs where
    facs = leftGreedyFactors n $ map brGenIdx gs
  
  replaceFac idxs = XDelta (-1) : map XSigma (_permutationBraid perm) where
    perm = (P.reversePermutation n) `P.multiply` (P.adjacentTranspositions n idxs)


-- | Replaces @sigma_i^-1@ generators by @(Delta^-1 * L_i)@.
replaceInversesNaive :: [BrGen] -> [XGen]
replaceInversesNaive gens = concatMap f gens where 
  f (Sigma    i) = [ XSigma i ]
  f (SigmaInv i) = [ XDelta (-1) , XL i ]

--------------------------------------------------------------------------------

-- | Temporary data structure to be used during the normal form computation
data XGen
  = XDelta !Int   -- ^ @Delta^k@
  | XSigma !Int   -- ^ @Sigma_j@
  | XL     !Int   -- ^ @L_j = Delta * sigma_j^-1@
  | XTauL  !Int   -- ^ @tau(L_j)@
  deriving (Eq,Show)

--  | XR     !Int   -- ^ @R_j = sigma_j^-1 * Delta@

isXDelta :: XGen -> Bool
isXDelta x = case x of { XDelta {} -> True ; _ -> False }

-- | We move the all @Delta@'s to the left
moveDeltasLeft :: Int -> [XGen] -> (Int,[XGen])
moveDeltasLeft n input = (finalExp, finalPosWord) where
  
  (XDelta finalExp : finalPosWord) =  reverse $ worker 0 (reverse input) 

  -- we start from the right end, and work towards the left end
  worker  dexp [] = [ XDelta dexp ]
  worker !dexp xs = this' ++ worker dexp' rest where 
    (delta,notdelta) = span isXDelta xs
    (this ,rest    ) = span (not . isXDelta) notdelta
    dexp' = dexp + sumDeltas delta
    this' = if even dexp' 
      then this
      else map xtau this

  sumDeltas :: [XGen] -> Int
  sumDeltas xs = foldl' (+) 0 [ k | XDelta k <- xs ]

  -- | The @X -> Delta^-1 * X * Delta@ inner automorphism
  xtau :: XGen -> XGen
  xtau (XSigma j) = XSigma (n-j)
  xtau (XDelta k) = XDelta k  
  xtau (XL     k) = XTauL  k  
  xtau (XTauL  k) = XL     k  

--------------------------------------------------------------------------------

expandPosXWord :: Int -> [XGen] -> [Int]
expandPosXWord n = concatMap f where

  posHalfTwist = _halfTwist n

  jtau :: Int -> Int
  jtau j = n-j

  posLTable    = listArray (1,n-1) [ _permutationBraid (posLPerm n i) | i<-[1..n-1] ] :: Array Int [Int]
  posTauLTable = amap (map jtau) posLTable

  -- posRTable = listArray (1,n-1) [ _permutationBraid (posRPerm n i) | i<-[1..n-1] ] :: Array Int [Int]

  f x = case x of
    XSigma i -> [i]
    XL     i -> posLTable    ! i
    XTauL  i -> posTauLTable ! i
    XDelta i 
      | i > 0     -> concat (replicate i posHalfTwist)
      | i < 0     -> error "expandPosXWord: negative delta power"
      | otherwise -> []

  -- word :: Braid n -> [Int]
  -- word (Braid gens) = map brGenIdx gens


-- | for debugging only
expandAnyXWord :: forall n. KnownNat n => [XGen] -> Braid n
expandAnyXWord xgens = braid where
  n = numberOfStrands braid

  braid = composeMany (map f xgens)

  posHalfTwist = halfTwist            :: Braid n
  negHalfTwist = inverse posHalfTwist :: Braid n

  posLTable    = listArray (1,n-1) [ permutationBraid (posLPerm n i) | i<-[1..n-1] ] :: Array Int (Braid n)
  posTauLTable = amap tau posLTable

  -- posRTable = listArray (1,n-1) [ permutationBraid (posRPerm n i) | i<-[1..n-1] ] :: Array Int (Braid n)

  f :: XGen -> Braid n
  f x = case x of
    XSigma i -> sigma i
    XL     i -> posLTable    ! i
    XTauL  i -> posTauLTable ! i
    XDelta i 
      | i > 0     -> composeMany (replicate   i  posHalfTwist)
      | i < 0     -> composeMany (replicate (-i) negHalfTwist)
      | otherwise -> identity



--------------------------------------------------------------------------------

-- | @posL k@ (denoted as @L_k@) is a /positive word/ which 
-- satisfies @Delta = L_k * sigma_k@, or:
-- 
-- > (inverse halfTwist) `compose` (posL k) ~=~ sigmaInv k@
-- 
-- Thus we can replace any word with a positive word plus some @Delta^-1@\'s
--
posL :: KnownNat n => Int -> Braid n
posL k = braid where
  n = numberOfStrands braid
  braid = permutationBraid (posLPerm n k)

-- | @posR k n@ (denoted as @R_k@) is a /permutation braid/ which 
-- satisfies @Delta = sigma_k * R_k@
-- 
-- > (posR k) `compose` (inverse halfTwist) ~=~ sigmaInv k@
-- 
-- Thus we can replace any word with a positive word plus some @Delta^-1@'s
--
posR :: KnownNat n => Int -> Braid n
posR k = braid where
  n = numberOfStrands braid
  braid = permutationBraid (posRPerm n k)

-- | The permutation @posL k :: Braid n@ is realizing
posLPerm :: Int -> Int -> Permutation
posLPerm n k 
  | k>0 && k<n  = (P.reversePermutation n `P.multiply` P.adjacentTransposition n k)
  | otherwise   = error "posLPerm: index out of range"

-- | The permutation @posR k :: Braid n@ is realizing
posRPerm :: Int -> Int -> Permutation
posRPerm n k 
  | k>0 && k<n  = (P.adjacentTransposition n k `P.multiply` P.reversePermutation n )
  | otherwise   = error "posRPerm: index out of range"

--------------------------------------------------------------------------------

-- | We recognize left-greedy factors which are @Delta@-s (easy, since they are the only ones
-- with length @(n choose 2)@), and move them to the left, returning their summed exponent
-- and the filtered new factors. We also filter trivial permutations (which should only happen 
-- for the trivial braid, but it happens there?)
filterDeltaFactors :: Int -> [[Int]] -> (Int, [[Int]])
filterDeltaFactors n facs = (exp',facs'') where

  (exp',facs') = go 0 (reverse facs)

  jtau j = n-j
  facs'' = reverse facs'
  maxlen = div (n*(n-1)) 2

  go !e []       = (e,[])
  go !e (xs:xxs)  
    | null xs             = go e xxs
    | length xs == maxlen = go (e+1) xxs
    | otherwise           =  
        if even e
          then let (e',yys) = go e xxs in (e' ,          xs : yys) 
          else let (e',yys) = go e xxs in (e' , map jtau xs : yys)  
      
-------------------------------------------------------------------------------- 
                                      
-- | We try to apply the braid relations to move elements left between
-- factor boundaries, to get something unique 
--
normalizeFactors :: [[Int]] -> [[Int]]
normalizeFactors = error "normalizeFactors: not implemented yet"
{-
normalizeFactors = reverse . go . reverse where
  go []  = []
  go [f] = [f]
  go (xs:rest@(ys:_)) = case xs of
    (a:b:c:_) | a == c && b == a+1  -> 
              | a == c && b == a-1  ->
              | otherwise           ->
-}

-------------------------------------------------------------------------------- 

-- | Given a /positive/ word, we apply left-greedy factorization of
-- that word into subwords representing /permutation braids/.
--
-- Example 5.1 from the above handbook:
--
-- > leftGreedyFactors 7 [1,3,2,2,1,3,3,2,3,2] == [[1,3,2],[2,1,3],[3,2,3],[2]]
--
leftGreedyFactors :: Int -> [Int] -> [[Int]]
leftGreedyFactors n input = runST (action input) where

  action :: forall s. [Int] -> ST s [[Int]]
  action input = do

    perm <- newArray_ (1,n) :: ST s (STUArray s Int Int)
    forM_ [1..n] $ \i -> writeArray perm i i
    let doSwap :: Int -> ST s ()
        doSwap i = do
          a <- readArray perm  i
          b <- readArray perm (i+1)
          writeArray perm  i    b
          writeArray perm (i+1) a
               
    mat <- newArray ((1,1),(n,n)) 0 :: ST s (STUArray s (Int,Int) Int)
    let clearMat = forM_ [1..n] $ \i -> 
          forM_ [1..n] $ \j -> writeArray mat (i,j) 0
          
    let doAdd1 :: Int -> Int -> ST s Int
        doAdd1 i j = do
          x <- readArray mat (i,j)
          let y = x+1
          writeArray mat (i,j) y 
          writeArray mat (j,i) y
          return y
           
    let worker :: [Int] -> ST s [[Int]]
        worker []     = return [[]]
        worker (p:ps) = do
          u <- readArray perm  p 
          v <- readArray perm (p+1)
          c <- doAdd1 u v 
          doSwap p
          if c<=1
            then do
              (f:fs) <- worker ps
              return ((p:f):fs)
            else do
              clearMat
              fs <- worker (p:ps)
              return ([]:fs)
              
    worker input

--------------------------------------------------------------------------------


--------------------------------------------------------------------------------

#ifdef QUICKCHECK

prop_braidnf_reduce :: KnownNat n => Braid n -> Bool
prop_braidnf_reduce braid = (braidNormalForm' braid == braidNormalForm braid)

prop_braidnf_reprs :: KnownNat n => Braid n -> Bool
prop_braidnf_reprs braid = (nf == nf') where
  nf  = braidNormalForm braid 
  nf' = braidNormalForm braid'
  braid' = nfReprWord nf

#endif

--------------------------------------------------------------------------------